"""P1 macro elements.

This element's definition appears in https://doi.org/10.1007/s00211-018-0970-6
(Christiansen, Hu, 2018)
"""

import typing

import sympy

from ..finite_element import CiarletElement
from ..functionals import IntegralAgainst, ListOfFunctionals, PointEvaluation
from ..functions import FunctionInput
from ..geometry import SetOfPoints
from ..piecewise_functions import PiecewiseFunction
from ..references import Reference
from ..symbols import x


class P1Macro(CiarletElement):
    """P1 macro finite element on a triangle."""

    def __init__(self, reference: Reference, order: int):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
        """
        third = sympy.Rational(1, 3)
        zero = sympy.Integer(0)
        one = sympy.Integer(1)
        tris: typing.List[SetOfPoints] = [
            ((zero, zero), (one, zero), (third, third)),
            ((one, zero), (zero, one), (third, third)),
            ((zero, one), (zero, zero), (third, third)),
        ]
        tris = [tuple(reference.get_point(p) for p in t) for t in tris]
        invmap = reference.get_inverse_map_to_self()
        poly: typing.List[FunctionInput] = [
            PiecewiseFunction({q: 1 for q in tris}, 2),
            PiecewiseFunction({q: x[0] for q in tris}, 2),
            PiecewiseFunction({q: x[1] for q in tris}, 2),
            PiecewiseFunction({
                tris[0]: 3 * invmap[1],
                tris[1]: 3 * (1 - invmap[0] - invmap[1]),
                tris[2]: 3 * invmap[0],
            }, 2)]

        dofs: ListOfFunctionals = []
        for v_n, v in enumerate(reference.vertices):
            dofs.append(PointEvaluation(reference, v, entity=(0, v_n)))
        dofs.append(IntegralAgainst(reference, 1, entity=(2, 0)))

        super().__init__(
            reference, order, poly, dofs, reference.tdim, 1
        )

    names = ["P1 macro"]
    references = ["triangle"]
    min_order = 1
    max_order = 1
    continuity = "C0"
    last_updated = "2023.06"



"""Brezzi-Douglas-Marini elements on simplices.

This element's definition appears in https://doi.org/10.1007/BF01389710
(Brezzi, Douglas, Marini, 1985)
"""

import typing

from ..finite_element import CiarletElement
from ..functionals import IntegralMoment, ListOfFunctionals, NormalIntegralMoment
from ..functions import FunctionInput
from ..moments import make_integral_moment_dofs
from ..polynomials import polynomial_set_vector
from ..references import NonDefaultReferenceError, Reference
from .lagrange import Lagrange
from .nedelec import NedelecFirstKind


class BDM(CiarletElement):
    """Brezzi-Douglas-Marini Hdiv finite element."""

    def __init__(self, reference: Reference, order: int, variant: str = "equispaced"):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
            variant: The variant of the element
        """
        if reference.vertices != reference.reference_vertices:
            raise NonDefaultReferenceError()

        poly: typing.List[FunctionInput] = []
        poly += polynomial_set_vector(reference.tdim, reference.tdim, order)

        dofs: ListOfFunctionals = make_integral_moment_dofs(
            reference,
            facets=(NormalIntegralMoment, Lagrange, order, {"variant": variant}),
            cells=(IntegralMoment, NedelecFirstKind, order - 1, {"variant": variant}),
        )
        self.variant = variant

        super().__init__(reference, order, poly, dofs, reference.tdim, reference.tdim)

    def init_kwargs(self) -> typing.Dict[str, typing.Any]:
        """Return the kwargs used to create this element.

        Returns:
            Keyword argument dictionary
        """
        return {"variant": self.variant}

    names = ["Brezzi-Douglas-Marini", "BDM", "N2div"]
    references = ["triangle", "tetrahedron"]
    min_order = 1
    continuity = "H(div)"
    last_updated = "2023.06"



"""Polynomial sets."""

import typing
from itertools import product

from ..functions import ScalarFunction, VectorFunction
from ..symbols import AxisVariablesNotSingle, x


def polynomial_set_1d(
    dim: int, order: int, variables: AxisVariablesNotSingle = x
) -> typing.List[ScalarFunction]:
    """One dimensional polynomial set.

    Args:
        dim: The number of variables
        order: The maximum polynomial degree
        variables: The variables to use

    Returns:
        A set of polynomials
    """
    if dim == 1:
        return [ScalarFunction(variables[0] ** i) for i in range(order + 1)]
    if dim == 2:
        return [
            ScalarFunction(variables[0] ** i * variables[1] ** j)
            for j in range(order + 1)
            for i in range(order + 1 - j)
        ]
    if dim == 3:
        return [
            ScalarFunction(variables[0] ** i * variables[1] ** j * variables[2] ** k)
            for k in range(order + 1)
            for j in range(order + 1 - k)
            for i in range(order + 1 - k - j)
        ]
    raise ValueError(f"Unsupported dimension: {dim}")


def polynomial_set_vector(
    domain_dim: int, range_dim: int, order: int, variables: AxisVariablesNotSingle = x
) -> typing.List[VectorFunction]:
    """Polynomial set.

    Args:
        domain_dim: The number of variables
        range_dim: The dimension of the output vector
        order: The maximum polynomial degree
        variables: The variables to use

    Returns:
        A set of polynomials
    """
    set1d = polynomial_set_1d(domain_dim, order, variables)
    return [
        VectorFunction([p if i == j else 0 for j in range(range_dim)])
        for p in set1d
        for i in range(range_dim)
    ]


def Hdiv_polynomials(
    domain_dim: int, range_dim: int, order: int, variables: AxisVariablesNotSingle = x
) -> typing.List[VectorFunction]:
    """Hdiv conforming polynomial set.

    Args:
        domain_dim: The number of variables
        range_dim: The dimension of the output vector
        order: The maximum polynomial degree
        variables: The variables to use

    Returns:
        A set of polynomials
    """
    assert domain_dim == range_dim
    if domain_dim == 2:
        return [VectorFunction((
            variables[0] * variables[0] ** (order - 1 - j) * variables[1] ** j,
            variables[1] * variables[0] ** (order - 1 - j) * variables[1] ** j,
        )) for j in range(order)]
    if domain_dim == 3:
        basis: typing.List[VectorFunction] = []
        for j in range(order):
            for k in range(order - j):
                p = variables[0] ** (order - 1 - j - k) * variables[1] ** j * variables[2] ** k
                basis.append(VectorFunction((variables[0] * p, variables[1] * p, variables[2] * p)))
        return basis

    raise ValueError(f"Unsupported dimension: {domain_dim}")


def Hcurl_polynomials(
    domain_dim: int, range_dim: int, order: int, variables: AxisVariablesNotSingle = x
) -> typing.List[VectorFunction]:
    """Hcurl conforming polynomial set.

    Args:
        domain_dim: The number of variables
        range_dim: The dimension of the output vector
        order: The maximum polynomial degree
        variables: The variables to use

    Returns:
        A set of polynomials
    """
    assert domain_dim == range_dim
    if domain_dim == 2:
        return [VectorFunction((
            variables[0] ** (order - 1 - j) * variables[1] ** (j + 1),
            -variables[0] ** (order - j) * variables[1] ** j,
        )) for j in range(order)]
    if domain_dim == 3:
        poly: typing.List[VectorFunction] = []
        poly += [VectorFunction((
            variables[0] ** (m - 1) * variables[1] ** n * variables[2] ** (order - m - n + 1),
            0, -variables[0] ** m * variables[1] ** n * variables[2] ** (order - m - n)
        )) for n in range(order) for m in range(1, order + 1 - n)]
        poly += [VectorFunction((
            0, variables[0] ** m * variables[1] ** (n - 1) * variables[2] ** (order - m - n + 1),
            -variables[0] ** m * variables[1] ** n * variables[2] ** (order - m - n)
        )) for m in range(order) for n in range(1, order + 1 - m)]
        poly += [VectorFunction((
            variables[0] ** (order - n) * variables[1] ** n,
            -variables[0] ** (order + 1 - n) * variables[1] ** (n - 1), 0
        )) for n in range(1, order + 1)]
        return poly
    raise ValueError(f"Unsupported dimension: {domain_dim}")


def quolynomial_set_1d(
    dim: int, order: int, variables: AxisVariablesNotSingle = x
) -> typing.List[ScalarFunction]:
    """One dimensional quolynomial set.

    Args:
        dim: The number of variables
        order: The maximum polynomial degree
        variables: The variables to use

    Returns:
        A set of polynomials
    """
    basis = []
    for j in product(range(order + 1), repeat=dim):
        poly = ScalarFunction(1)
        for a, b in zip(variables, j):
            poly *= a ** b
        basis.append(poly)
    return basis


def quolynomial_set_vector(
    domain_dim: int, range_dim: int, order: int, variables: AxisVariablesNotSingle = x
) -> typing.List[VectorFunction]:
    """Quolynomial set.

    Args:
        domain_dim: The number of variables
        range_dim: The dimension of the output vector
        order: The maximum polynomial degree
        variables: The variables to use

    Returns:
        A set of polynomials
    """
    set1d = quolynomial_set_1d(domain_dim, order, variables)
    return [
        VectorFunction([p if i == j else 0 for j in range(range_dim)])
        for p in set1d
        for i in range(range_dim)
    ]


def Hdiv_quolynomials(
    domain_dim: int, range_dim: int, order: int, variables: AxisVariablesNotSingle = x
) -> typing.List[VectorFunction]:
    """Hdiv conforming quolynomial set.

    Args:
        domain_dim: The number of variables
        range_dim: The dimension of the output vector
        order: The maximum polynomial degree
        variables: The variables to use

    Returns:
        A set of polynomials
    """
    assert domain_dim == range_dim
    basis: typing.List[VectorFunction] = []
    for d in range(domain_dim):
        for j in product(range(order), repeat=domain_dim - 1):
            poly = 1
            for a, b in zip(variables, j[:d] + (order,) + j[d:]):
                poly *= a ** b
            basis.append(VectorFunction([poly if i == d else 0 for i in range(domain_dim)]))
    return basis


def Hcurl_quolynomials(
    domain_dim: int, range_dim: int, order: int, variables: AxisVariablesNotSingle = x
) -> typing.List[VectorFunction]:
    """Hcurl conforming quolynomial set.

    Args:
        domain_dim: The number of variables
        range_dim: The dimension of the output vector
        order: The maximum polynomial degree
        variables: The variables to use

    Returns:
        A set of polynomials
    """
    assert domain_dim == range_dim
    basis: typing.List[VectorFunction] = []
    for d in range(domain_dim):
        for j in product(
            *[range(order) if i == d else range(order + 1) for i in range(domain_dim)]
        ):
            if order not in j:
                continue
            poly = 1
            for a, b in zip(variables, j):
                poly *= a ** b
            basis.append(VectorFunction([poly if i == d else 0 for i in range(domain_dim)]))
    return basis


def serendipity_indices(
    total: int, linear: int, dim: int, done: typing.Optional[typing.List[int]] = None
) -> typing.List[typing.List[int]]:
    """Get the set indices for a serendipity polynomial set.

    Args:
        dim: The number of variables
        order: The maximum polynomial degree
        variables: The variables to use

    Returns:
        A set of polynomials
    """
    if done is None:
        done = []
    if len(done) == dim:
        if done.count(1) >= linear:
            return [done]
        return []
    if len(done) == dim - 1:
        return serendipity_indices(total, linear, dim, done=done + [total - sum(done)])
    out = []
    for i in range(total - sum(done) + 1):
        out += serendipity_indices(total, linear, dim, done + [i])
    return out


def serendipity_set_1d(
    dim: int, order: int, variables: AxisVariablesNotSingle = x
) -> typing.List[ScalarFunction]:
    """One dimensional serendipity set.

    Args:
        dim: The number of variables
        order: The maximum polynomial degree
        variables: The variables to use

    Returns:
        A set of polynomials
    """
    basis: typing.List[ScalarFunction] = []
    for s in range(order + 1, order + dim + 1):
        for i in serendipity_indices(s, s - order, dim):
            p = 1
            for j, k in zip(variables, i):
                p *= j ** k
            basis.append(ScalarFunction(p))
    return basis


def serendipity_set_vector(
    domain_dim: int, range_dim: int, order: int, variables: AxisVariablesNotSingle = x
) -> typing.List[VectorFunction]:
    """Serendipity set.

    Args:
        domain_dim: The number of variables
        range_dim: The dimension of the output vector
        order: The maximum polynomial degree
        variables: The variables to use

    Returns:
        A set of polynomials
    """
    set1d = serendipity_set_1d(domain_dim, order, variables)
    return [
        VectorFunction([p if i == j else 0 for j in range(range_dim)])
        for p in set1d
        for i in range(range_dim)
    ]


def Hdiv_serendipity(
    domain_dim: int, range_dim: int, order: int, variables: AxisVariablesNotSingle = x
) -> typing.List[VectorFunction]:
    """Hdiv conforming serendipity set.

    Args:
        domain_dim: The number of variables
        range_dim: The dimension of the output vector
        order: The maximum polynomial degree
        variables: The variables to use

    Returns:
        A set of polynomials
    """
    assert domain_dim == range_dim
    if domain_dim == 2:
        return [
            VectorFunction((variables[0] ** (order + 1),
                            (order + 1) * variables[0] ** order * variables[1])),
            VectorFunction(((order + 1) * variables[0] * variables[1] ** order,
                            variables[1] ** (order + 1))),
        ]
    if domain_dim == 3:
        a = []
        if order == 0:
            a.append(VectorFunction((0, variables[0] * variables[2], -variables[0] * variables[1])))
            a.append(VectorFunction((variables[1] * variables[2], 0, -variables[0] * variables[1])))
        else:
            for i in range(order + 1):
                p = variables[1] ** i * variables[2] ** (order - i)
                a.append(VectorFunction((0, variables[0] * variables[2] * p,
                                         -variables[0] * variables[1] * p)))

                p = variables[0] ** i * variables[2] ** (order - i)
                a.append(VectorFunction((variables[1] * variables[2] * p, 0,
                                         -variables[0] * variables[1] * p)))

                p = variables[0] ** i * variables[1] ** (order - i)
                a.append(VectorFunction((variables[1] * variables[2] * p,
                                         -variables[0] * variables[2] * p, 0)))

        return [i.curl() for i in a]

    raise NotImplementedError()


def Hcurl_serendipity(
    domain_dim: int, range_dim: int, order: int, variables: AxisVariablesNotSingle = x
) -> typing.List[VectorFunction]:
    """Hcurl conforming serendipity set.

    Args:
        domain_dim: The number of variables
        range_dim: The dimension of the output vector
        order: The maximum polynomial degree
        variables: The variables to use

    Returns:
        A set of polynomials
    """
    assert domain_dim == range_dim
    if domain_dim == 2:
        return [
            VectorFunction(((order + 1) * variables[0] ** order * variables[1],
                            -variables[0] ** (order + 1))),
            VectorFunction((variables[1] ** (order + 1),
                            (order + 1) * variables[0] * variables[1] ** order)),
        ]
    if domain_dim == 3:
        out: typing.List[VectorFunction] = []
        if order == 1:
            out += [
                VectorFunction((0, variables[0] * variables[2], -variables[0] * variables[1])),
                VectorFunction((variables[1] * variables[2], 0, -variables[0] * variables[1])),
            ]
        else:
            for i in range(order):
                p = variables[0] ** i * variables[2] ** (order - 1 - i)
                out.append(VectorFunction((variables[1] * variables[2] * p, 0,
                                           -variables[0] * variables[1] * p)))

                p = variables[1] ** i * variables[2] ** (order - 1 - i)
                out.append(VectorFunction((0, variables[0] * variables[2] * p,
                                           -variables[0] * variables[1] * p)))

                p = variables[0] ** i * variables[1] ** (order - 1 - i)
                out.append(VectorFunction((variables[1] * variables[2] * p,
                                           -variables[0] * variables[2] * p, 0)))

        for p in serendipity_set_1d(domain_dim, order + 1):
            out.append(VectorFunction(tuple(p.diff(i) for i in variables)))
        return out

    raise NotImplementedError()


def prism_polynomial_set_1d(
    dim: int, order: int, variables: AxisVariablesNotSingle = x
) -> typing.List[ScalarFunction]:
    """One dimensional polynomial set.

    Args:
        dim: The number of variables
        order: The maximum polynomial degree
        variables: The variables to use

    Returns:
        A set of polynomials
    """
    assert dim == 3
    return [
        ScalarFunction(variables[0] ** i * variables[1] ** j * variables[2] ** k)
        for k in range(order + 1)
        for j in range(order + 1)
        for i in range(order + 1 - j)
    ]


def prism_polynomial_set_vector(
    domain_dim: int, range_dim: int, order: int, variables: AxisVariablesNotSingle = x
) -> typing.List[VectorFunction]:
    """Polynomial set for a prism.

    Args:
        domain_dim: The number of variables
        range_dim: The dimension of the output vector
        order: The maximum polynomial degree
        variables: The variables to use

    Returns:
        A set of polynomials
    """
    set1d = prism_polynomial_set_1d(domain_dim, order, variables)
    return [
        VectorFunction([p if i == j else 0 for j in range(range_dim)])
        for p in set1d
        for i in range(range_dim)
    ]


def pyramid_polynomial_set_1d(
    dim: int, order: int, variables: AxisVariablesNotSingle = x
) -> typing.List[ScalarFunction]:
    """One dimensional polynomial set.

    Args:
        dim: The number of variables
        order: The maximum polynomial degree
        variables: The variables to use

    Returns:
        A set of polynomials
    """
    assert dim == 3
    if order == 0:
        return [ScalarFunction(1)]

    poly = polynomial_set_1d(3, order)
    for d in range(order):
        for i in range(d + 1):
            for j in range(d + 1 - i):
                p = variables[0] ** i * variables[1] ** j * variables[2] ** (d - i - j)
                p *= (variables[0] * variables[1] / (1 - variables[2])) ** (order - d)
                poly.append(ScalarFunction(p))

    return poly


def pyramid_polynomial_set_vector(
    domain_dim: int, range_dim: int, order: int, variables: AxisVariablesNotSingle = x
) -> typing.List[VectorFunction]:
    """Polynomial set for a pyramid.

    Args:
        domain_dim: The number of variables
        range_dim: The dimension of the output vector
        order: The maximum polynomial degree
        variables: The variables to use

    Returns:
        A set of polynomials
    """
    set1d = pyramid_polynomial_set_1d(domain_dim, order, variables)
    return [
        VectorFunction([p if i == j else 0 for j in range(range_dim)])
        for p in set1d
        for i in range(range_dim)
    ]



"""Reduced Hsieh-Clough-Tocher elements on simplices.

This element's definition appears in https://doi.org/10.2307/2006147
(Ciarlet, 1978)
"""

import typing

import sympy

from ..finite_element import CiarletElement
from ..functionals import DerivativePointEvaluation, ListOfFunctionals, PointEvaluation
from ..functions import FunctionInput, ScalarFunction
from ..piecewise_functions import PiecewiseFunction
from ..references import NonDefaultReferenceError, Reference
from ..symbols import x


class ReducedHsiehCloughTocher(CiarletElement):
    """Reduced Hsieh-Clough-Tocher finite element."""

    def __init__(self, reference: Reference, order: int):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
        """
        assert order == 3
        assert reference.name == "triangle"
        if reference.vertices != reference.reference_vertices:
            raise NonDefaultReferenceError()

        dofs: ListOfFunctionals = []
        for v_n, vs in enumerate(reference.vertices):
            dofs.append(PointEvaluation(reference, vs, entity=(0, v_n)))
            dofs.append(DerivativePointEvaluation(reference, vs, (1, 0), entity=(0, v_n)))
            dofs.append(DerivativePointEvaluation(reference, vs, (0, 1), entity=(0, v_n)))

        mid = tuple(sympy.Rational(sum(i), len(i)) for i in zip(*reference.vertices))

        subs = [
            (reference.vertices[0], reference.vertices[1], mid),
            (reference.vertices[1], reference.vertices[2], mid),
            (reference.vertices[2], reference.vertices[0], mid)]

        piece_list = [tuple(ScalarFunction(p) for _ in range(3))
                      for p in [1, x[0], x[1], x[0]**2, x[0]*x[1], x[1]**2,
                                x[0]**3 - x[1]**3]]
        piece_list.append((
            ScalarFunction(4*x[0]**3 - 3*x[0]*x[1]**2 + 2*x[0]*x[1] + 4*x[1]**2),
            ScalarFunction(7*x[0]**3 + 12*x[0]**2*x[1] - 7*x[0]**2 + 9*x[0]*x[1]**2
                           - 14*x[0]*x[1] + 5*x[0] + 4*x[1] - 1),
            ScalarFunction(3*x[0]**3 + x[0]**2 - 2*x[1]**3 + 5*x[1]**2)))
        piece_list.append((
            ScalarFunction(25*x[0]**3 - 24*x[0]*x[1]**2 + 30*x[0]*x[1] - 24*x[1]**2),
            ScalarFunction(35*x[0]**3 + 33*x[0]**2*x[1] - 21*x[0]**2 - 12*x[0]*x[1]**2
                           + 12*x[0] - 28*x[1]**3 - 3*x[1] - 1),
            ScalarFunction(3*x[0]**3 + 21*x[0]**2*x[1] + 15*x[0]**2 - 23*x[1]**3 - 9*x[1]**2)))

        poly: typing.List[FunctionInput] = []
        poly += [
            PiecewiseFunction({i: j for i, j in zip(subs, p)}, 2)
            for p in piece_list]
        poly = reference.map_polyset_from_default(poly)

        super().__init__(
            reference, order, poly, dofs, reference.tdim, 1
        )

    names = ["reduced Hsieh-Clough-Tocher", "rHCT"]
    references = ["triangle"]
    min_order = 3
    max_order = 3
    # continuity = "C1"
    continuity = "C0"
    last_updated = "2023.06"



import pytest

import symfem
from symfem.mappings import MappingNotImplemented
from symfem.references import NonDefaultReferenceError
from symfem.utils import allequal

from .utils import test_elements


@pytest.mark.parametrize(
    ("cell_type", "element_type", "order", "kwargs"),
    [[reference, element, order, kwargs]
     for reference, i in test_elements.items() for element, j in i.items()
     for kwargs, k in j for order in k])
def test_push_forward(
    elements_to_test, cells_to_test, cell_type, element_type, order, kwargs,
    speed
):
    if elements_to_test != "ALL" and element_type not in elements_to_test:
        pytest.skip()
    if cells_to_test != "ALL" and cell_type not in cells_to_test:
        pytest.skip()
    if speed == "fast":
        if order > 2:
            pytest.skip()
        if order == 2 and cell_type in ["tetrahedron", "hexahedron", "prism", "pyramid"]:
            pytest.skip()

    if cell_type == "interval":
        vertices = [(3, ), (1, )]
    elif cell_type == "triangle":
        vertices = [(1, 1), (2, 2), (1, 4)]
    elif cell_type == "quadrilateral":
        vertices = [(1, 1), (2, 2), (1, 3), (2, 4)]
    elif cell_type == "tetrahedron":
        vertices = [(1, 1, 1), (2, 2, 2), (-1, 3, 2), (4, 0, 0)]
    elif cell_type == "hexahedron":
        vertices = [(1, 1, 1), (2, 2, 2), (-1, 3, 2), (0, 4, 3),
                    (4, 0, 0), (5, 1, 1), (2, 2, 1), (3, 3, 2)]
    elif cell_type == "prism":
        vertices = [(1, 1, 1), (2, 2, 1), (1, 4, 2), (0, 1, 1), (1, 2, 1), (0, 4, 2)]
    elif cell_type == "pyramid":
        vertices = [(1, 1, 0), (2, 2, 0), (1, 3, 1), (2, 4, 1), (-1, -1, -1)]

    else:
        raise ValueError(f"Unsupported cell type: {cell_type}")

    try:
        e2 = symfem.create_element(cell_type, element_type, order, vertices=vertices, **kwargs)
    except NonDefaultReferenceError:
        pytest.xfail("Cannot create element on non-default reference.")
    e = symfem.create_element(cell_type, element_type, order, **kwargs)

    try:
        assert allequal(e.map_to_cell(vertices), e2.get_basis_functions())
    except MappingNotImplemented:
        pytest.xfail("Mapping not implemented for this element.")



"""Test Bell elements."""

import symfem
from symfem.symbols import t, x


def test_bell_polyset():
    b = symfem.create_element("triangle", "Bell", 5)
    for p in b.get_polynomial_basis():
        gradp = [p.diff(x[0]), p.diff(x[1])]
        for en in range(b.reference.sub_entity_count(1)):
            edge = b.reference.sub_entity(1, en)
            variables = [o + sum(a[i] * t[0] for a in edge.axes)
                         for i, o in enumerate(edge.origin)]
            n = edge.normal()
            normal_deriv = gradp[0] * n[0] + gradp[1] * n[1]
            normal_deriv = normal_deriv.subs(x, variables)
            assert normal_deriv.diff(t[0]).diff(t[0]).diff(t[0]).diff(t[0]) == 0



"""Test Hsieh-Clough-Tocher elements."""

import sympy

import symfem
from symfem.symbols import t, x
from symfem.utils import allequal

half = sympy.Rational(1, 2)


def test_continuity():
    e = symfem.create_element("triangle", "Alfeld-Sorokina", 2)
    for f in e.get_polynomial_basis():
        # edge from (1,0) to (1/3,1/3)
        f1 = f.get_piece((half, 0))
        f2 = f.get_piece((half, half))
        div_f1 = f1.div()
        div_f2 = f2.div()
        line = ((1 - 2 * t[0], t[0]))
        f1 = f1.subs(x[:2], line)
        f2 = f2.subs(x[:2], line)
        div_f1 = div_f1.subs(x[:2], line)
        div_f2 = div_f2.subs(x[:2], line)
        assert allequal(f1, f2)
        assert allequal(div_f1, div_f2)

        # edge from (0,1) to (1/3,1/3)
        f1 = f.get_piece((half, half))
        f2 = f.get_piece((0, half))
        div_f1 = f1.div()
        div_f2 = f2.div()
        line = ((t[0], 1 - 2 * t[0]))
        f1 = f1.subs(x[:2], line)
        f2 = f2.subs(x[:2], line)
        div_f1 = div_f1.subs(x[:2], line)
        div_f2 = div_f2.subs(x[:2], line)
        assert allequal(f1, f2)
        assert allequal(div_f1, div_f2)

        # edge from (0,0) to (1/3,1/3)
        f1 = f.get_piece((0, half))
        f2 = f.get_piece((half, 0))
        div_f1 = f1.div()
        div_f2 = f2.div()
        line = ((t[0], t[0]))
        f1 = f1.subs(x[:2], line)
        f2 = f2.subs(x[:2], line)
        div_f1 = div_f1.subs(x[:2], line)
        div_f2 = div_f2.subs(x[:2], line)
        assert allequal(f1, f2)
        assert allequal(div_f1, div_f2)



"""Hellan-Herrmann-Johnson elements on simplices.

This element's definition appears in https://arxiv.org/abs/1909.09687
(Arnold, Walker, 2020)

For an alternative construction see (Sinwel, 2009) and sections 4.4.2.2 and 4.4.3.2
https://numa.jku.at/media/filer_public/b7/42/b74263c9-f723-4076-b1b2-c2726126bf32/phd-sinwel.pdf
"""

import typing

from ..finite_element import CiarletElement
from ..functionals import IntegralMoment, ListOfFunctionals, NormalInnerProductIntegralMoment
from ..functions import FunctionInput, MatrixFunction
from ..moments import make_integral_moment_dofs
from ..polynomials import polynomial_set_vector
from ..references import NonDefaultReferenceError, Reference
from .lagrange import Lagrange


class HellanHerrmannJohnson(CiarletElement):
    """A Hellan-Herrmann-Johnson element."""

    def __init__(self, reference: Reference, order: int, variant: str = "equispaced"):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
            variant: The variant of the element
        """
        if reference.vertices != reference.reference_vertices:
            raise NonDefaultReferenceError()
        assert reference.name == "triangle" or reference.name == "tetrahedron"

        poly: typing.List[FunctionInput] = []
        directions: typing.List[typing.Tuple[typing.Tuple[int, ...], ...]] = []
        directions_extra: typing.List[typing.Tuple[typing.Tuple[int, ...], ...]] = []

        if reference.tdim == 2:
            poly = [((p[0], p[1]), (p[1], p[2]))
                    for p in polynomial_set_vector(reference.tdim, 3, order)]
            directions = [((0, 1), (1, 0)),
                          ((-2, 1), (1, 0)),
                          ((0, -1), (-1, 2))]
            directions_extra = []
        if reference.tdim == 3:
            poly = [((p[0], p[1], p[2]), (p[1], p[3], p[4]), (p[2], p[4], p[5]))
                    for p in polynomial_set_vector(reference.tdim, 6, order)]
            directions = [((-2, 1, 0), (1, 0, 0), (0, 0, 0)),
                          ((0, 1, -1), (1, -2, 1), (-1, 1, 0)),
                          ((0, 0, 0), (0, 0, -1), (0, -1, 2)),
                          ((0, 0, 1), (0, 0, 0), (1, 0, 0))]
            directions_extra = [((0, 0, -1), (0, 0, 1), (-1, 1, 0)),
                                ((0, -1, 0), (-1, 0, 1), (0, 1, 0))]

        dofs: ListOfFunctionals = make_integral_moment_dofs(
            reference,
            facets=(NormalInnerProductIntegralMoment, Lagrange, order,
                    {"variant": variant}),
        )

        # cell functions
        if order > 0:
            space = Lagrange(reference, order - 1, variant)
            basis = space.get_basis_functions()
            for p, dof in zip(basis, space.dofs):
                for d in directions:
                    dofs.append(IntegralMoment(
                        reference, p * MatrixFunction(d), dof, entity=(reference.tdim, 0),
                        mapping="double_covariant"))
        # cell functions extra
        space_extra = Lagrange(reference, order, variant)
        basis_extra = space_extra.get_basis_functions()
        for p, dof in zip(basis_extra, space_extra.dofs):
            for d in directions_extra:
                dofs.append(IntegralMoment(
                    reference, p * MatrixFunction(d), dof, entity=(reference.tdim, 0),
                    mapping="double_covariant"))

        self.variant = variant

        super().__init__(reference, order, poly, dofs, reference.tdim, reference.tdim ** 2,
                         (reference.tdim, reference.tdim))

    def init_kwargs(self) -> typing.Dict[str, typing.Any]:
        """Return the kwargs used to create this element.

        Returns:
            Keyword argument dictionary
        """
        return {"variant": self.variant}

    names = ["Hellan-Herrmann-Johnson", "HHJ"]
    references = ["triangle", "tetrahedron"]
    min_order = 0
    continuity = "inner H(div)"
    last_updated = "2023.08"



import pytest
import sympy

import symfem
from symfem.utils import allequal


def test_cache():
    e = symfem.create_element("triangle", "Lagrange", 2)

    m1 = e.get_dual_matrix(caching=False, inverse=True)

    def a(*args, **kwargs):
        raise RuntimeError()

    oldinv = sympy.Matrix.inv
    sympy.Matrix.inv = a

    with pytest.raises(RuntimeError):
        e.get_dual_matrix(caching=False, inverse=True)

    m2 = e.get_dual_matrix(inverse=True)

    sympy.Matrix.inv = oldinv

    assert allequal(m1, m2)



"""Alfeld-Sorokina element on a triangle.

This element's definition appears in https://doi.org/10.1007/s10543-015-0557-x
(Alfeld, Sorokina, 2015)
"""

import typing

import sympy

from ..finite_element import CiarletElement
from ..functionals import DotPointEvaluation, ListOfFunctionals, PointDivergenceEvaluation
from ..functions import FunctionInput, VectorFunction
from ..piecewise_functions import PiecewiseFunction
from ..references import NonDefaultReferenceError, Reference
from ..symbols import x


class AlfeldSorokina(CiarletElement):
    """Alfeld-Sorokina finite element."""

    def __init__(self, reference: Reference, order: int):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
        """
        assert order == 2
        assert reference.name == "triangle"
        if reference.vertices != reference.reference_vertices:
            raise NonDefaultReferenceError()

        dofs: ListOfFunctionals = []
        for v_n, vs in enumerate(reference.vertices):
            dofs.append(DotPointEvaluation(reference, vs, (1, 0), entity=(0, v_n)))
            dofs.append(DotPointEvaluation(reference, vs, (0, 1), entity=(0, v_n)))
            dofs.append(PointDivergenceEvaluation(reference, vs, entity=(0, v_n)))

        for e_n in range(reference.sub_entity_count(1)):
            sub_ref = reference.sub_entity(1, e_n)
            dofs.append(DotPointEvaluation(reference, sub_ref.midpoint(), (1, 0), entity=(1, e_n)))
            dofs.append(DotPointEvaluation(reference, sub_ref.midpoint(), (0, 1), entity=(1, e_n)))

        mid = tuple(sympy.Rational(sum(i), len(i)) for i in zip(*reference.vertices))

        subs = [
            (reference.vertices[0], reference.vertices[1], mid),
            (reference.vertices[1], reference.vertices[2], mid),
            (reference.vertices[2], reference.vertices[0], mid)]

        piece_list = [tuple(VectorFunction((p, 0)) for _ in range(3))
                      for p in [1, x[0], x[1], x[0]**2, x[0]*x[1], x[1]**2]]
        piece_list += [tuple(VectorFunction((0, p)) for _ in range(3))
                       for p in [1, x[0], x[1], x[0]**2, x[0]*x[1], x[1]**2]]

        piece_list.append((
            VectorFunction((4*x[0] + 12*x[1]**2 - 4*x[1], 6*x[0]*x[1] - 4*x[1])),
            VectorFunction((3*x[0]**2 + 2*x[0] + 4*x[1] - 1, -3*x[0]**2 + 4*x[0] - 2*x[1] - 1)),
            VectorFunction((3*x[0]**2 + 6*x[0]*x[1] + 3*x[1]**2, 9*x[0]**2 - 4*x[0] - 3*x[1]**2))))
        piece_list.append((
            VectorFunction((144*x[0]*x[1] + 10*x[0] + 12*x[1]**2 - 10*x[1],
                            -30*x[0]*x[1] + 36*x[1]**2 - 10*x[1])),
            VectorFunction((-69*x[0]**2 + 104*x[0] + 46*x[1] - 25,
                            24*x[0]**2 - 26*x[0] + 4*x[1] + 2)),
            VectorFunction((39*x[0]**2 + 96*x[0]*x[1] + 21*x[1]**2,
                            -10*x[0] + 6*x[1]**2))))
        piece_list.append((
            VectorFunction((-14*x[0] + 12*x[1]**2 + 14*x[1],
                            42*x[0]*x[1] - 108*x[1]**2 + 14*x[1])),
            VectorFunction((3*x[0]**2 - 16*x[0] + 22*x[1] - 1,
                            24*x[0]**2 + 144*x[0]*x[1] - 50*x[0] - 92*x[1] + 26)),
            VectorFunction((-33*x[0]**2 + 24*x[0]*x[1] + 21*x[1]**2,
                            14*x[0] - 66*x[1]**2))))

        poly: typing.List[FunctionInput] = []
        poly += [
            PiecewiseFunction({i: j for i, j in zip(subs, p)}, 2)
            for p in piece_list]

        super().__init__(
            reference, order, poly, dofs, reference.tdim, 1
        )

    names = ["Alfeld-Sorokina", "AS"]
    references = ["triangle"]
    min_order = 2
    max_order = 2
    continuity = "C0"
    last_updated = "2023.05"



"""Test Hsieh-Clough-Tocher elements."""

import pytest
import sympy

import symfem
from symfem.functions import ScalarFunction
from symfem.symbols import t, x
from symfem.utils import allequal

half = sympy.Rational(1, 2)


@pytest.mark.parametrize("family", ["HCT", "rHCT"])
def test_c1_continuity(family):
    e = symfem.create_element("triangle", family, 3)
    for f in e.get_polynomial_basis():
        # edge from (1,0) to (1/3,1/3)
        f1 = f.get_piece((half, 0))
        f2 = f.get_piece((half, half))
        grad_f1 = f1.grad(2)
        grad_f2 = f2.grad(2)
        line = ((1 - 2 * t[0], t[0]))
        f1 = f1.subs(x[:2], line)
        f2 = f2.subs(x[:2], line)
        grad_f1 = grad_f1.subs(x[:2], line)
        grad_f2 = grad_f2.subs(x[:2], line)
        assert allequal(f1, f2)
        assert allequal(grad_f1, grad_f2)

        # edge from (0,1) to (1/3,1/3)
        f1 = f.get_piece((half, half))
        f2 = f.get_piece((0, half))
        grad_f1 = f1.grad(2)
        grad_f2 = f2.grad(2)
        line = ((t[0], 1 - 2 * t[0]))
        f1 = f1.subs(x[:2], line)
        f2 = f2.subs(x[:2], line)
        grad_f1 = grad_f1.subs(x[:2], line)
        grad_f2 = grad_f2.subs(x[:2], line)
        assert allequal(f1, f2)
        assert allequal(grad_f1, grad_f2)

        # edge from (0,0) to (1/3,1/3)
        f1 = f.get_piece((0, half))
        f2 = f.get_piece((half, 0))
        grad_f1 = f1.grad(2)
        grad_f2 = f2.grad(2)
        line = ((t[0], t[0]))
        f1 = f1.subs(x[:2], line)
        f2 = f2.subs(x[:2], line)
        grad_f1 = grad_f1.subs(x[:2], line)
        grad_f2 = grad_f2.subs(x[:2], line)
        assert allequal(f1, f2)
        assert allequal(grad_f1, grad_f2)


def test_rcht_linear_normal_derivatices():
    e = symfem.create_element("triangle", "rHCT", 3)
    for f in e.get_polynomial_basis():
        # Check that normal derivatives are linear
        f1 = f.get_piece((half, 0)).diff(x[1]).subs(x[1], 0)
        f2 = f.get_piece((half, half))
        f2 = (f2.diff(x[0]) + f2.diff(x[1])).subs(x[1], 1 - x[0])
        f3 = f.get_piece((0, half)).diff(x[0]).subs(x[0], 0)
        assert f1.diff(x[0]).diff(x[0]) == 0
        assert f2.diff(x[0]).diff(x[0]) == 0
        assert f3.diff(x[1]).diff(x[1]) == 0


def test_rhct_integral():
    element = symfem.create_element("triangle", "rHCT", 3)
    ref = element.reference
    f1 = element.get_basis_function(1).directional_derivative((1, 0))
    f2 = element.get_basis_function(6).directional_derivative((1, 0))

    integrand = f1 * f2
    third = sympy.Rational(1, 3)
    integrand.pieces[((0, 0), (1, 0), (third, third))] = ScalarFunction(0)
    integrand.pieces[((1, 0), (0, 1), (third, third))] = ScalarFunction(0)

    expr = integrand.pieces[((0, 1), (0, 0), (third, third))].as_sympy()
    assert len(integrand.pieces) == 3

    assert sympy.integrate(sympy.integrate(
        expr, (x[1], x[0], 1 - 2 * x[0])), (x[0], 0, third)) == integrand.integral(ref, x)



"""Test Hellan-Herrmann-Johnson element."""

import pytest

from symfem import create_element
from symfem.functions import VectorFunction
from symfem.symbols import x
from symfem.utils import allequal


@pytest.mark.parametrize('reference', ['triangle', 'tetrahedron'])
@pytest.mark.parametrize('order', [0, 1, 2])
def test_create(reference, order):

    element = create_element(reference, "HHJ", order)

    # Get the basis functions associated with the interior
    basis = element.get_basis_functions()
    functions = [basis[d] for d in element.entity_dofs(element.reference.tdim, 0)]

    if reference == "triangle":
        # Check that these tensor functions have zero normal normal component on each edge
        for f in functions:
            M, n = f.subs(x[0], 1 - x[1]), VectorFunction((1, 1))
            assert allequal((M @ n).dot(n), 0)
            M, n = f.subs(x[0], 0), VectorFunction((-1, 0))
            assert allequal((M @ n).dot(n), 0)
            M, n = f.subs(x[1], 0), VectorFunction((0, -1))
            assert allequal((M @ n).dot(n), 0)

    if reference == "tetrahedron":
        # Check that these tensor functions have zero normal normal component on each face
        for f in functions:
            M, n = f.subs(x[0], 1 - x[1] - x[2]), VectorFunction((1, 1, 1))
            assert allequal((M @ n).dot(n), 0)
            M, n = f.subs(x[0], 0), VectorFunction((-1, 0, 0))
            assert allequal((M @ n).dot(n), 0)
            M, n = f.subs(x[1], 0), VectorFunction((0, -1, 0))
            assert allequal((M @ n).dot(n), 0)
            M, n = f.subs(x[2], 0), VectorFunction((0, 0, -1))
            assert allequal((M @ n).dot(n), 0)



import pytest

import symfem


@pytest.mark.parametrize("cell, order", [
    (c, o) for c, m in [("interval", 5), ("triangle", 4), ("quadrilateral", 4), ("tetrahedron", 3),
                        ("hexahedron", 3), ("prism", 3), ("pyramid", 3)] for o in range(m)])
def test_legendre(cell, order):
    basis = symfem.polynomials.orthonormal_basis(cell, order, 0)[0]
    e = symfem.create_element(cell, "P", order, variant="legendre")
    assert symfem.utils.allequal(e.get_basis_functions(), basis)


@pytest.mark.parametrize("cell, order", [
    (c, o) for c, m in [("interval", 5), ("quadrilateral", 4), ("hexahedron", 3)]
    for o in range(m)])
def test_lobatto(cell, order):
    basis = symfem.polynomials.lobatto_basis(cell, order, False)
    e = symfem.create_element(cell, "P", order, variant="lobatto")
    assert symfem.utils.allequal(e.get_basis_functions()[-len(basis):], basis)



"""Dual polynomials."""

import typing

import sympy

from ..functions import ScalarFunction


def l2_dual(cell: str, poly: typing.List[ScalarFunction]) -> typing.List[ScalarFunction]:
    """Compute the L2 dual of a set of polynomials.

    Args:
        cell: The cell type
        poly: The set of polynomial

    Returns:
        The L2 dual polynomials
    """
    from ..create import create_reference
    reference = create_reference(cell)

    matrix = sympy.Matrix([[(p * q).integral(reference) for q in poly] for p in poly])
    minv = matrix.inv("LU")

    out = []
    for i in range(minv.rows):
        f = ScalarFunction(0)
        for j, p in zip(minv.row(i), poly):
            f += j * p
        out.append(f)

    return out



"""Functions to map functions between cells."""

import sympy

from .functions import AnyFunction, FunctionInput, MatrixFunction, parse_function_input
from .geometry import PointType
from .symbols import x


class MappingNotImplemented(NotImplementedError):
    """Exception thrown when a mapping is not implemented for an element."""


def identity(
    f_in: FunctionInput, map: PointType, inverse_map: PointType,
    substitute: bool = True,
) -> AnyFunction:
    """Map functions.

    Args:
        f_in: The function
        map: The map from the reference cell to the physical cell
        inverse_map: The map to the reference cell from the physical cell
        substitute: Should the inverse map be substituted in?

    Returns:
        The mapped function
    """
    f = parse_function_input(f_in)
    if substitute:
        f = f.subs(x, inverse_map)
    return f


def l2(
    f_in: FunctionInput, map: PointType, inverse_map: PointType,
    substitute: bool = True,
) -> AnyFunction:
    """Map functions, scaling by the determinant of the jacobian.

    Args:
        f_in: The function
        map: The map from the reference cell to the physical cell
        inverse_map: The map to the reference cell from the physical cell
        substitute: Should the inverse map be substituted in?

    Returns:
        The mapped function
    """
    tdim = len(inverse_map)
    jdet = MatrixFunction([[i.diff(x[j]) for j in range(tdim)] for i in map]).det().as_sympy()
    assert isinstance(jdet, sympy.core.expr.Expr)
    f = parse_function_input(f_in)
    if substitute:
        f = f.subs(x, inverse_map)
    return f / abs(jdet)


def covariant(
    f_in: FunctionInput, map: PointType, inverse_map: PointType,
    substitute: bool = True,
) -> AnyFunction:
    """Map H(curl) functions.

    Args:
        f_in: The function
        map: The map from the reference cell to the physical cell
        inverse_map: The map to the reference cell from the physical cell
        substitute: Should the inverse map be substituted in?

    Returns:
        The mapped function
    """
    tdim = len(inverse_map)
    f = parse_function_input(f_in)
    if substitute:
        f = f.subs(x, inverse_map)

    assert f.is_vector

    j_inv = MatrixFunction([[i.diff(x[j]) for i in inverse_map] for j in range(tdim)])
    return j_inv @ f


def contravariant(
    f_in: FunctionInput, map: PointType, inverse_map: PointType,
    substitute: bool = True
) -> AnyFunction:
    """Map H(div) functions.

    Args:
        f_in: The function
        map: The map from the reference cell to the physical cell
        inverse_map: The map to the reference cell from the physical cell
        substitute: Should the inverse map be substituted in?

    Returns:
        The mapped function
    """
    tdim = len(inverse_map)
    f = parse_function_input(f_in)
    if substitute:
        f = f.subs(x, inverse_map)

    assert f.is_vector

    jacobian = MatrixFunction([[i.diff(x[j]) for j in range(tdim)] for i in map])
    jacobian /= jacobian.det()
    return jacobian @ f


def double_covariant(
    f_in: FunctionInput, map: PointType, inverse_map: PointType,
    substitute: bool = True,
) -> MatrixFunction:
    """Map matrix functions.

    Args:
        f_in: The function
        map: The map from the reference cell to the physical cell
        inverse_map: The map to the reference cell from the physical cell
        substitute: Should the inverse map be substituted in?

    Returns:
        The mapped function
    """
    tdim = len(inverse_map)
    f = parse_function_input(f_in)
    if substitute:
        f = f.subs(x, inverse_map)
    assert f.is_matrix

    j_inv = MatrixFunction([[i.diff(x[j]) for i in inverse_map] for j in range(tdim)])
    return j_inv @ f @ j_inv.transpose()


def double_contravariant(
    f_in: FunctionInput, map: PointType, inverse_map: PointType,
    substitute: bool = True,
) -> MatrixFunction:
    """Map matrix functions.

    Args:
        f_in: The function
        map: The map from the reference cell to the physical cell
        inverse_map: The map to the reference cell from the physical cell
        substitute: Should the inverse map be substituted in?

    Returns:
        The mapped function
    """
    tdim = len(inverse_map)
    f = parse_function_input(f_in)
    if substitute:
        f = f.subs(x, inverse_map)
    assert f.is_matrix

    jacobian = MatrixFunction([[i.diff(x[j]) for j in range(tdim)] for i in map])
    jacobian /= jacobian.det()
    return jacobian @ f @ jacobian.transpose()


def identity_inverse_transpose(
    f_in: FunctionInput, map: PointType, inverse_map: PointType,
    substitute: bool = True,
) -> AnyFunction:
    """Inverse transpose of identity().

    Args:
        f_in: The function
        map: The map from the reference cell to the physical cell
        inverse_map: The map to the reference cell from the physical cell
        substitute: Should the inverse map be substituted in?

    Returns:
        The mapped function
    """
    f = parse_function_input(f_in)
    if substitute:
        f = f.subs(x, inverse_map)
    return f


def l2_inverse_transpose(
    f_in: FunctionInput, map: PointType, inverse_map: PointType,
    substitute: bool = True,
) -> AnyFunction:
    """Inverse transpose of l2().

    Args:
        f_in: The function
        map: The map from the reference cell to the physical cell
        inverse_map: The map to the reference cell from the physical cell
        substitute: Should the inverse map be substituted in?

    Returns:
        The mapped function
    """
    tdim = len(inverse_map)
    f = parse_function_input(f_in)
    if substitute:
        f = f.subs(x, inverse_map)
    jdet = MatrixFunction([[i.diff(x[j]) for j in range(tdim)] for i in map]).det().as_sympy()
    assert isinstance(jdet, sympy.core.expr.Expr)
    return f * abs(jdet)


def covariant_inverse_transpose(
    f_in: FunctionInput, map: PointType, inverse_map: PointType,
    substitute: bool = True,
) -> AnyFunction:
    """Inverse transpose of covariant().

    Args:
        f_in: The function
        map: The map from the reference cell to the physical cell
        inverse_map: The map to the reference cell from the physical cell
        substitute: Should the inverse map be substituted in?

    Returns:
        The mapped function
    """
    tdim = len(inverse_map)
    f = parse_function_input(f_in)
    if substitute:
        f = f.subs(x, inverse_map)

    assert f.is_vector

    jacobian = MatrixFunction([[i.diff(x[j]) for j in range(tdim)] for i in map])
    return jacobian @ f


def contravariant_inverse_transpose(
    f_in: FunctionInput, map: PointType, inverse_map: PointType,
    substitute: bool = True,
) -> AnyFunction:
    """Inverse transpose of contravariant().

    Args:
        f_in: The function
        map: The map from the reference cell to the physical cell
        inverse_map: The map to the reference cell from the physical cell
        substitute: Should the inverse map be substituted in?

    Returns:
        The mapped function
    """
    tdim = len(inverse_map)
    f = parse_function_input(f_in)
    if substitute:
        f = f.subs(x, inverse_map)

    assert f.is_vector

    j_inv = MatrixFunction([[i.diff(x[j]) for i in inverse_map] for j in range(tdim)])
    j_inv /= j_inv.det()
    return j_inv @ f


def identity_inverse(
    f_in: FunctionInput, map: PointType, inverse_map: PointType,
    substitute: bool = True,
) -> AnyFunction:
    """Inverse of identity().

    Args:
        f_in: The function
        map: The map from the reference cell to the physical cell
        inverse_map: The map to the reference cell from the physical cell
        substitute: Should the inverse map be substituted in?

    Returns:
        The mapped function
    """
    return identity(f_in, inverse_map, map, substitute)


def l2_inverse(
    f_in: FunctionInput, map: PointType, inverse_map: PointType,
    substitute: bool = True,
) -> AnyFunction:
    """Inverse of l2().

    Args:
        f_in: The function
        map: The map from the reference cell to the physical cell
        inverse_map: The map to the reference cell from the physical cell
        substitute: Should the inverse map be substituted in?

    Returns:
        The mapped function
    """
    return l2(f_in, inverse_map, map, substitute)


def covariant_inverse(
    f_in: FunctionInput, map: PointType, inverse_map: PointType,
    substitute: bool = True,
) -> AnyFunction:
    """Inverse of covariant().

    Args:
        f_in: The function
        map: The map from the reference cell to the physical cell
        inverse_map: The map to the reference cell from the physical cell
        substitute: Should the inverse map be substituted in?

    Returns:
        The mapped function
    """
    return covariant(f_in, inverse_map, map, substitute)


def contravariant_inverse(
    f_in: FunctionInput, map: PointType, inverse_map: PointType,
    substitute: bool = True
) -> AnyFunction:
    """Inverse of contravariant().

    Args:
        f_in: The function
        map: The map from the reference cell to the physical cell
        inverse_map: The map to the reference cell from the physical cell
        substitute: Should the inverse map be substituted in?

    Returns:
        The mapped function
    """
    return contravariant(f_in, inverse_map, map, substitute)


def double_covariant_inverse(
    f_in: FunctionInput, map: PointType, inverse_map: PointType,
    substitute: bool = True,
) -> AnyFunction:
    """Inverse of double_covariant().

    Args:
        f_in: The function
        map: The map from the reference cell to the physical cell
        inverse_map: The map to the reference cell from the physical cell
        substitute: Should the inverse map be substituted in?

    Returns:
        The mapped function
    """
    return double_covariant(f_in, inverse_map, map, substitute)


def double_contravariant_inverse(
    f_in: FunctionInput, map: PointType, inverse_map: PointType,
    substitute: bool = True
) -> AnyFunction:
    """Inverse of double_contravariant().

    Args:
        f_in: The function
        map: The map from the reference cell to the physical cell
        inverse_map: The map to the reference cell from the physical cell
        substitute: Should the inverse map be substituted in?

    Returns:
        The mapped function
    """
    return double_contravariant(f_in, inverse_map, map, substitute)
